/* gtkbutlerwavebuffer.c: class that stores data in a butlerbuffer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <gtk/gtk.h>
#include "gtkbutlerwavebuffer.h"


static void gtk_butler_wave_buffer_class_init (GtkButlerWaveBufferClass *klass);
static void gtk_butler_wave_buffer_init       (GtkButlerWaveBuffer      *butlerbuffer);
static gboolean gtk_butler_wave_buffer_is_editable     (GtkWaveBuffer *wavebuffer);
static guint32 gtk_butler_wave_buffer_get_rate         (GtkWaveBuffer *wavebuffer);
static GtkWaveBufferType gtk_butler_wave_buffer_get_datatype  (GtkWaveBuffer *wavebuffer);
static guint32 gtk_butler_wave_buffer_get_length       (GtkWaveBuffer *wavebuffer);
static guint32 gtk_butler_wave_buffer_get_num_channels (GtkWaveBuffer *wavebuffer);
static void gtk_butler_wave_buffer_get_samples         (GtkWaveBuffer *wavebuffer,
                                                          guint32 start,
                                                          guint32 length,
                                                          guint32 channel_mask,
                                                          gpointer data);
static gint gtk_butler_wave_buffer_set_samples         (GtkEditableWaveBuffer *editable,
                                                          guint32 start,
                                                          guint32 length,
                                                          guint32 channel_mask,
                                                          gpointer data);
static gint gtk_butler_wave_buffer_insert              (GtkEditableWaveBuffer *editable,
                                                          guint32 start,
                                                          guint32 length);
static gint gtk_butler_wave_buffer_delete              (GtkEditableWaveBuffer *editable,
                                                      guint32 start,
                                                      guint32 length);



GtkType
gtk_butler_wave_buffer_get_type (void)
{
  static GtkType butler_type = 0;

  if (!butler_type)
    {
      static const GtkTypeInfo butler_info =
      {
        "GtkButlerWaveBuffer",
        sizeof (GtkButlerWaveBuffer),
        sizeof (GtkButlerWaveBufferClass),
        (GtkClassInitFunc) gtk_butler_wave_buffer_class_init,
        (GtkObjectInitFunc) gtk_butler_wave_buffer_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      butler_type = gtk_type_unique (GTK_TYPE_EDITABLE_WAVE_BUFFER, &butler_info);
    }

  return butler_type;
}

static void
gtk_butler_wave_buffer_finalize (GtkObject *obj)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (obj);

  g_print ("Finalize called.\n");
  fclose (butlerbuffer->file);

  if (butlerbuffer->in_use != NULL)
    g_mutex_free (butlerbuffer->in_use);
}

static void
gtk_butler_wave_buffer_class_init (GtkButlerWaveBufferClass *klass)
{
  GtkObjectClass *object_class;
  GtkWaveBufferClass *wavebuffer_class;
  GtkEditableWaveBufferClass *editable_class;

  object_class = (GtkObjectClass *) klass;
  wavebuffer_class = (GtkWaveBufferClass *) klass;
  editable_class = (GtkEditableWaveBufferClass *) klass;

  object_class->finalize = gtk_butler_wave_buffer_finalize;

  wavebuffer_class->get_rate = gtk_butler_wave_buffer_get_rate;
  wavebuffer_class->get_datatype = gtk_butler_wave_buffer_get_datatype;
  wavebuffer_class->get_length = gtk_butler_wave_buffer_get_length;
  wavebuffer_class->get_num_channels = gtk_butler_wave_buffer_get_num_channels;
  wavebuffer_class->get_samples = gtk_butler_wave_buffer_get_samples;
  editable_class->set_samples = gtk_butler_wave_buffer_set_samples;
  editable_class->insert = gtk_butler_wave_buffer_insert;
  editable_class->delete = gtk_butler_wave_buffer_delete;
}

static void
gtk_butler_wave_buffer_init (GtkButlerWaveBuffer *butlerbuffer)
{
  butlerbuffer->n_channels = 0;
  butlerbuffer->d_type = GTK_WAVE_BUFFER_TYPE_NULL;
  butlerbuffer->length = 0;
  butlerbuffer->rate = 0;
  butlerbuffer->file = tmpfile ();

  if (g_threads_got_initialized)
    butlerbuffer->in_use = g_mutex_new ();
  else
    butlerbuffer->in_use = NULL;
}

static guint32
gtk_butler_wave_buffer_get_rate (GtkWaveBuffer *wavebuffer)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (wavebuffer);

  return butlerbuffer->rate;
}

static GtkWaveBufferType
gtk_butler_wave_buffer_get_datatype (GtkWaveBuffer *wavebuffer)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (wavebuffer);

  return butlerbuffer->d_type;
}

static guint32
gtk_butler_wave_buffer_get_length (GtkWaveBuffer *wavebuffer)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (wavebuffer);

  return butlerbuffer->length;
}

static guint32
gtk_butler_wave_buffer_get_num_channels (GtkWaveBuffer *wavebuffer)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (wavebuffer);

  return butlerbuffer->n_channels;
}

static void
gtk_butler_wave_buffer_get_samples (GtkWaveBuffer *wavebuffer,
                                      guint32 start,
                                      guint32 length,
                                      guint32 channel_mask,
                                      gpointer data)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (wavebuffer);
  guint32 mask, i, j, width;

  width = gtk_wave_buffer_type_width (butlerbuffer->d_type);

  mask = (1 << (butlerbuffer->n_channels + 1)) - 1;

  if ((mask & channel_mask) == mask) {
	  /* get all channels */

  } else {
      guint8 junk[width * butlerbuffer->n_channels];

      for (i = 0; i < length; i++)
        for (j = 0; j < butlerbuffer->n_channels; j++)
          if (channel_mask & (1 << j))
            {
              fread (data, width, 1, butlerbuffer->file);
              data = (gpointer) (((guint8*) data) + width);
            }
          else
            fread (junk, width, 1, butlerbuffer->file);
    }

  if (butlerbuffer->in_use != NULL)
    g_mutex_unlock (butlerbuffer->in_use);
}

static gint
gtk_butler_wave_buffer_set_samples (GtkEditableWaveBuffer *editable,
                                      guint32 start,
                                      guint32 length,
                                      guint32 channel_mask,
                                      gpointer data)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (editable);
  guint32 mask, i, j, width;

  if (butlerbuffer->in_use != NULL)
    g_mutex_lock (butlerbuffer->in_use);

  width = gtk_wave_buffer_type_width (butlerbuffer->d_type);
  fseek (butlerbuffer->file, start * width * butlerbuffer->n_channels, SEEK_SET);

  mask = (1 << (butlerbuffer->n_channels + 1)) - 1;

  if ((mask & channel_mask) == mask)
    fwrite (data, width * butlerbuffer->n_channels, length, butlerbuffer->file);
  else
    {
      guint8 junk[width * butlerbuffer->n_channels];

      for (i = 0; i < length; i++)
        for (j = 0; j < butlerbuffer->n_channels; j++)
          if (channel_mask & (1 << j))
            {
              fwrite (data, width, 1, butlerbuffer->file);
              data = (gpointer) (((guint8*) data) + width);
            }
          else
            fread (junk, width, 1, butlerbuffer->file);
    }

  gtk_editable_wave_buffer_set_modified (editable, start, length);

  if (butlerbuffer->in_use != NULL)
    g_mutex_unlock (butlerbuffer->in_use);

  return 0;
}

static void
gtk_butler_move_data (GtkButlerWaveBuffer *butlerbuffer,
                        guint32 from_pos,
                        guint32 to_pos,
                        guint32 length)
{
  guint8 block[4096];
  gint32 width, pos, size;

  if (from_pos == to_pos)
    return;

  width = gtk_wave_buffer_type_width (butlerbuffer->d_type) * butlerbuffer->n_channels;

  length *= width;
  from_pos *= width;
  to_pos *= width;

  if (to_pos < from_pos)
    {
      pos = 0;
      while (pos < length)
        {
          size = MIN (4096, length - pos);

          fseek (butlerbuffer->file, from_pos + pos, SEEK_SET);
          fread (block, 1, size, butlerbuffer->file);

          fseek (butlerbuffer->file, to_pos + pos, SEEK_SET);
          fwrite (block, 1, size, butlerbuffer->file);

          pos += size;
        }
    }
  else
    {
      pos = length;
      while (pos > 0)
        {
          size = MIN (4096, pos);

          fseek (butlerbuffer->file, from_pos + pos, SEEK_SET);
          fread (block, 1, size, butlerbuffer->file);

          fseek (butlerbuffer->file, to_pos + pos, SEEK_SET);
          fwrite (block, 1, size, butlerbuffer->file);

          pos -= size;
        }
    }
}

static gint
gtk_butler_wave_buffer_insert (GtkEditableWaveBuffer *editable,
                                 guint32 start,
                                 guint32 ins_length)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (editable);
  gint32 length;
  GRange range;

  length = gtk_wave_buffer_get_length (GTK_WAVE_BUFFER (editable));

  if (butlerbuffer->in_use != NULL)
    g_mutex_lock (butlerbuffer->in_use);

  gtk_butler_move_data (butlerbuffer, start, start + ins_length, length - start);
  butlerbuffer->length += ins_length;

  g_range_set (&range, start, length);
  gtk_signal_emit (GTK_OBJECT (editable), insert_data_signal, &range);

  gtk_editable_wave_buffer_set_modified (editable, start, butlerbuffer->length - start);

  if (butlerbuffer->in_use != NULL)
    g_mutex_unlock (butlerbuffer->in_use);

  return ferror (butlerbuffer->file);
}

static gint
gtk_butler_wave_buffer_delete (GtkEditableWaveBuffer *editable,
                                 guint32 start,
                                 guint32 length)
{
  GtkButlerWaveBuffer *butlerbuffer = GTK_BUTLER_WAVE_BUFFER (editable);
  guint32 orig_length;
  GRange range;

  if (butlerbuffer->in_use != NULL)
    g_mutex_lock (butlerbuffer->in_use);

  orig_length = butlerbuffer->length;
  gtk_butler_move_data (butlerbuffer, start + length, start, orig_length - length);

  butlerbuffer->length = orig_length - length;
  ftruncate (fileno (butlerbuffer->file), butlerbuffer->length);

  g_range_set (&range, start, length);
  gtk_signal_emit (GTK_OBJECT (editable), delete_data_signal, &range);

  gtk_editable_wave_buffer_set_modified (editable, start, orig_length - start);

  if (butlerbuffer->in_use != NULL)
    g_mutex_unlock (butlerbuffer->in_use);

  return ferror (butlerbuffer->file);
}

GtkObject *
gtk_butler_wave_buffer_new (guint16 n_channels,
                              GtkWaveBufferType d_type,
                              guint32 rate)
{
  GtkButlerWaveBuffer *wavebuffer;

  wavebuffer = gtk_type_new (GTK_TYPE_BUTLER_WAVE_BUFFER);
  wavebuffer->n_channels = n_channels;
  wavebuffer->d_type = d_type;
  wavebuffer->rate = rate;
  wavebuffer->length = 0;

  return GTK_OBJECT (wavebuffer);
}
