/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: seq.h,v 1.7 2000/11/04 16:35:47 pbd Exp $
*/

#ifndef __sequence_h__
#define __sequence_h__

#include <glib.h>

#ifdef __cplusplus
extern "C" {
#endif

/* XXX We can't include ardour.h here, because its C++, but this
   typedef MUST match the one for ARDOUR::Sample
*/

typedef float buffer_data_t;

typedef enum {
	ConstBuffer = 'C',
	FileBuffer = 'F'
} BufferType;

struct _piece;

typedef struct _buffer {
    /* return the type of this buffer */

    BufferType (*type)(struct _buffer *);

    /* this returns a string that will allow a future instantiation
       of the EDL to find the buffer again. For example, a file buffer
       class might return a full pathname.
    */

    gchar *(*id)(struct _buffer *);

    /* returns the number of items in the `buffer' */

    guint32 (*length) (struct _buffer *self);

    /* Transfers `cnt' buffer_data_t's from the buffer, starting at
       `start, to `dst'.
    */

    guint32 (*read) (struct _buffer *self, buffer_data_t *dst, guint32 start, guint32 cnt);

    struct _buffer *(*clone)(struct _buffer *self);

    /* this appends `cnt' items from `src', beginning at `start', to
       the end of `dst'. this is used EXCLUSIVELY when creating a
       new buffer for the sequence_clone() operation.
    */

    guint32 (*append) (struct _buffer *dst, struct _buffer *src, guint32 start, guint32 cnt);

    /* this attempts to put new sample data at the end of the buffer;
       it returns a new piece describing the successfully inserted
       data, or NULL if the insertion failed.
    */

    struct _piece *(*append_data) (struct _buffer *dst, buffer_data_t *, guint32 cnt);

    /* this is called when we free the buffer */

    void (*close)(struct _buffer *self);

    gpointer *private_data;
    guint32   refcnt;

} buffer_t;

buffer_t * buffer_new (BufferType (*type)(buffer_t *),
		       gchar *(*id)(buffer_t *),
		       guint32 (*len)(), 
		       guint32 (*read)(buffer_t *, buffer_data_t *, guint32, guint32),
		       buffer_t *(*clone)(buffer_t *),
		       guint32 (*append)(buffer_t *, buffer_t *, guint32, guint32),
		       struct _piece *(*append_data)(buffer_t *, buffer_data_t *, guint32),
		       void (*close)(buffer_t *),
		       gpointer data);

typedef struct _piece
{
    buffer_t *buffer;
    guint32 start;
    guint32 length;
    guint32 refcnt;
} piece_t;

#define piece_size(p) ((p)->length * sizeof (buffer_data_t))

piece_t *piece_new (buffer_t *b, guint32 start, guint32 cnt);

typedef struct _sequence {
    GList       *pieces;
    GList       *state_ptr;
    guint32      length;
    const gchar *path;        /* the EDL file */
    const gchar *backup_path; /* backup EDL file, used between saves */
    gboolean     loading;
    buffer_t    *default_buffer;     /* based on the base file */
    float        gain;
} sequence_t;

typedef enum {
	SequenceInsert,
        SequenceReplace,		
        SequenceDelete,
	SequenceSetGain
} SequenceOpType;

typedef struct _sequence_op {
    piece_t       *piece;
    SequenceOpType type;
    guint32        start;
    guint32        cnt;
    float          gain;
} sequence_op_t;

typedef struct _sequence_state {
    GList   *pieces;
    guint32  length;
    float    gain;
    sequence_op_t *op;
} sequence_state_t;

typedef gint (*SequenceFunction)(buffer_t *, guint32, guint32, gpointer);

sequence_t *sequence_new (const gchar *base_file, buffer_t *existing_buffer);
void        sequence_free (sequence_t *);

sequence_t *sequence_subsequence (sequence_t *, guint32 start, guint32 cnt);
sequence_t *sequence_copy (sequence_t *);

gint    sequence_apply (sequence_t *, guint32, guint32, SequenceFunction function, gpointer arg);
guint32 sequence_read (sequence_t *, buffer_data_t *buf, guint32 start, guint32 cnt);

int     sequence_insert_sequence (sequence_t *, guint32 where, sequence_t *);
int     sequence_insert_piece (sequence_t *, guint32 where, piece_t *);

int     sequence_delete (sequence_t *, guint32 start, guint32 length);

int     sequence_replace_data (sequence_t *, guint32 start, buffer_data_t *, guint32 cnt);
int     sequence_replace_piece (sequence_t *, guint32 start, piece_t *);
int     sequence_replace_sequence (sequence_t *, guint32 start, sequence_t *);

int     sequence_undo (sequence_t *);
int     sequence_redo (sequence_t *);

gint    sequence_state_stack_position (sequence_t *);
gint    sequence_set_state_stack_position (sequence_t *, guint pos);

void    sequence_set_gain (sequence_t *, float gain);

/* If `path' is NULL, save the state of the sequence to the default
   path, otherwise to `path'.
*/

gint    sequence_save (sequence_t *, const gchar *path);

#ifdef __cplusplus
}  /* extern "C" */
#endif

#endif  /* __sequence_h__ */
