/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: track.cc,v 1.56 2000/11/03 19:45:39 pbd Exp $
*/

#include <algobase.h>

#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <asm/msr.h>

#include <pbd/error.h>

#include "track.h"
#include "riff.h"
#include "alsa_device.h"
#include "utils.h"
#include "configuration.h"

#include "seq.h"
#include "filebuffer.h"
#include "constbuffer.h"

static buffer_t *silent_buffer = 0;

using namespace ARDOUR;

#undef TRACE_TIMING

#include <stdio.h>

Track::Track (Tape &t, 
	      ALSA::MultiChannelDevice *mcdp,
	      size_t tn, 
	      size_t bufcnt)

	: tape (t),
	  mcd (mcdp),
	  track_num (tn)

{
	size_t x;
	char buf[64];
	size_t bufsize;

	_ok = false;
	fd = -1;
	edl = 0;

	sprintf (buf, "Track %d", track_num + 1);
	_tagname = buf;
	
	_input_channel = -1;
	_output_channel = -1;
	atomic_set (&_record_enabled, 0);
	_monitor_input = 0;
	was_recording = false;
	was_rolling = false;
	xfade_sample = 0;
	_delay = 0;
	delay_left = 0;
	_muted = false;

	switch (mcd->sample_format()) {
	case SND_PCM_SFMT_S32_LE:
		if (Config->max_level) {
			max_level = Config->max_level;
		} else {
			max_level = INT_MAX;
		}
		break;

	case SND_PCM_SFMT_S16_LE:
		if (Config->max_level) {
			max_level = Config->max_level;
		} else {
			max_level = SHRT_MAX;
		}
		break;
	}

	nbufs = bufcnt;
	playback_buffers = new Buffer[nbufs];
	capture_buffers = new Buffer[nbufs];

	bufsize = Config->minimum_disk_io_bytes / sizeof (ARDOUR::Sample);

	for (x = 0; x < nbufs; x++) {
		playback_buffers[x].set_size (bufsize);
		capture_buffers[x].set_size (bufsize);
		playback_buffers[x].needs_refill = true;
	}

	playback_sample = 0;

	if (use_take (1)) {
		return;
	}

	_ok = true;
}

Track::~Track ()

{
	if (fd >= 0) {
		close (fd);
	}
}

int
Track::use_take (size_t take)

{
	if (load_take (take) < 0) {
		return -1;
	}
	take_changed ();
	return 0;
}

int
Track::load_take (size_t take)

{
	string path;
	struct stat statbuf;
	char buf[8];
	bool new_file = false;

	sprintf (buf, "%d.%d", track_num + 1, take);

	path = tape.path();
	path += '/';
	path += buf;

	tape.wait_for_butler ();

	if (fd > 0) {
		close (fd);
		if (sequence_save (edl, NULL)) {
			error << "Track " << track_num
			      << ": cannot save old EDL."
			      << endmsg;
			return -1;
		}
	}

	_current_take = 0;
	_path = "";

	if (stat (path.c_str(), &statbuf)) {
		if (errno == ENOENT) {
			new_file = true;
		} else {
			error << "Track " << track_num
			      << ": cannot stat file " << path << " ("
			      << strerror (errno)
			      << ')'
			      << endmsg;
			return -1;
		}
	}

	if ((fd = open (path.c_str(), O_RDWR|O_CREAT, 0644)) < 0) {
		error << "Track " << track_num
		      << ": cannot setup audio file " << path << " ("
		      << strerror (errno)
		      << ')'
		      << endmsg;
		return -1;
	}


	/* load the EDL */

	if ((edl = sequence_new (path.c_str(), 0)) == 0) {
		error << "Track " << track_num
		      << ": cannot load EDL."
		      << endmsg;
		return -1;
	}

	if (new_file) {
		if (silent_buffer == 0) {
			silent_buffer = const_buffer_new (0);
		}
		piece_t *piece = piece_new (silent_buffer, 0, tape.frames());
		sequence_insert_piece (edl, 0, piece);
		sequence_save (edl, NULL);
	}

	seek (playback_sample);

	_current_take = take;
	_path = path;
	return fd;
}

int
Track::do_xfade (size_t offset, size_t samples, bool recording)

{
	if (recording) {
		mcd->xfade_in_from_channel (_input_channel,
					    capture_ptr,
					    samples,
					    offset,
					    xfade_sample);
	} else {
		mcd->xfade_out_from_channel (_input_channel,
					     capture_ptr,
					     samples,
					     offset,
					     xfade_sample);
	}

	if (samples > xfade_sample) {
		xfade_sample = 0;
	} else {
		xfade_sample -= samples;
	}

	return 0;
}

int
Track::no_roll (size_t nsamples, bool tape_state_changing)
	
{
	_peak_power = 0;
	
	if (tape_state_changing) {
		return 0;
	}

	if (tape.get_auto_input() && _monitor_input) {
		mcd->request_monitor_input (_input_channel, true);
	} 

	/* do our own monitoring so that we can copy from the input
	   channel to the output channel. if we left this for the MCD,
	   it would only copy a channel to itself, which is not the
	   only possible scenario here.
	*/
	
	if (_monitor_input && _input_channel >= 0 && _output_channel >= 0) {
		mcd->copy_channel (_input_channel,
				   _output_channel,
				   nsamples, 0);
	}

	return 0;
}

int
Track::roll (size_t nsamples, bool recording_enabled, 
	     bool recording_is_legal, bool &need_butler)

{
	size_t this_samples;
	size_t samples_left;
	size_t offset;
	bool recording = false;
	bool i_am_rec_enabled = atomic_read (&_record_enabled);
	int inchan;
	int outchan;
	bool was_xfading;

	samples_left = nsamples;
	offset = 0;

	/* provide some measure of atomicity for the channels
	 */
	   
	inchan = _input_channel;
	outchan = _output_channel;

	if (recording_is_legal && recording_enabled && 
	    i_am_rec_enabled && inchan >= 0) {
		recording = true;
	} 

	was_xfading = xfade_sample;

	if (recording != was_recording) {
		xfade_sample = mcd->xfade_samples () - 1;
	}

	if (recording) {
		mcd->request_monitor_input (inchan, true);
	} else {
		if (tape.get_auto_input()) {
			mcd->request_monitor_input (inchan, false);
		} 
	}

	was_recording = recording;

	while (samples_left) {

		if (recording || xfade_sample) {
			this_samples = min (samples_left, capture_samples_left);
		} else {
			this_samples = min (samples_left, playback_samples_left);
		}

		if (xfade_sample) {
			do_xfade (offset, this_samples, recording);
			if (!mcd->hw_monitoring() && outchan >= 0) {
				mcd->write_to_channel (outchan,
						       capture_ptr,
						       this_samples,
						       offset);
			}

			capture_buffers[capture_buffer].dirty += this_samples;

		} else {
			if (recording) {

				/* Not in a cross-fade situation. Just
				   grab data from the channel and stuff
				   it away.
				*/
				
				mcd->read_from_channel (inchan, 
							capture_ptr,
							this_samples,
							offset);

				if (!mcd->hw_monitoring() && outchan >= 0) {
					mcd->write_to_channel (outchan,
							       capture_ptr,
							       this_samples,
							       offset);
				}

				capture_buffers[capture_buffer].dirty += this_samples;

			} else if (outchan >= 0) {

				/* playback situation */

				if (_monitor_input && !tape.get_auto_input()) {

					if (!_muted) {
						mcd->copy_channel (inchan, outchan,
								   this_samples, offset);
					}

				} else {
					if (!_muted) {
						mcd->write_to_channel (outchan, 
								       playback_ptr,
								       this_samples, 
								       offset);
					}
				}
			} 
		}

		if (capture_buffers[playback_buffer].needs_flush) {
			error << "Track " << track_num
			      << ": capture buffer " << playback_buffer
			      << " not yet flushed! Halting."
			      << endmsg;
			return -1;
		}

		/* Move the data from the playback buffer into the
		   capture buffer so that its primed.
		*/

		memcpy (capture_buffers[playback_buffer].buf + 
			(playback_ptr - playback_buffers[playback_buffer].buf),
			playback_ptr, this_samples * sizeof (ARDOUR::Sample));
		capture_buffers[playback_buffer].file_sample = 
			playback_buffers[playback_buffer].file_sample;

		samples_left -= this_samples;
		offset += this_samples;
		playback_samples_left -= this_samples;
		
		if ((!_monitor_input && !recording) || tape.get_auto_input()) {
			compute_peak<ARDOUR::Sample> (playback_ptr, &_peak_power, this_samples, 1, max_level);
		}

		playback_ptr += this_samples;

		if (recording_is_legal ||
		    capture_buffers[capture_buffer].dirty) {
			capture_samples_left -= this_samples;
			capture_ptr += this_samples;
		}

		if (playback_samples_left == 0) {
			playback_buffers[playback_buffer].needs_refill = true;
			need_butler = true;

			playback_buffer++;
			playback_buffer %= nbufs;

			if (playback_buffers[playback_buffer].needs_refill) {
				error << "Tape \""
				      << tape.name ()
				      << "\": Track " << track_num
				      << ": butler/buffer UNDERRUN at buf #"
				      << playback_buffer
				      << endmsg;
				return -1;
			}

			playback_samples_left = 
				playback_buffers[playback_buffer].samples;
			playback_ptr = playback_buffers[playback_buffer].buf;
		}

		if (capture_samples_left == 0) {
			if (capture_buffers[capture_buffer].dirty) {
				need_butler = true;
				capture_buffers[capture_buffer].needs_flush = true;
			}
			capture_buffer++;
			capture_buffer %= nbufs;

			if (capture_buffers[capture_buffer].needs_flush) {
				error << "Tape \""
				      << tape.name()
				      << "\": Track " << track_num
				      << ": butler/buffer OVERRUN at buf #"
				      << capture_buffer
				      << endmsg;
				return -1;
			}
			
			capture_samples_left = 
				capture_buffers[capture_buffer].samples;
			capture_ptr = capture_buffers[capture_buffer].buf;
		}
	}

	playback_sample += nsamples;

	int ret;

	if (recording) {
		if (was_xfading) {
			/* we were already doing an xfade.
			   if it was finished, then return 0 (track
			   can stop). return 1 otherwise (track
			   must continue).
			*/

			ret = xfade_sample ? 1 : 0;
		} else {
			/* recording, but we weren't xfading, so there
			   is a stop necessary, including a possible
			   xfade.
			*/

			ret = 1;
		}

	} else {
		/* if we didn't recording anything, then we can stop
		   immediately.
		*/
		
		ret = 0;
	}

	return ret;
}

int
Track::seek (size_t which_sample)

{
	if (which_sample > _delay) {
		playback_sample = which_sample - _delay;
		delay_left = 0;
	} else {
		playback_sample = 0;
		delay_left = 0;
	}

	file_sample = playback_sample;

	if (invalidate_buffers (false)) {
		return -1;
	}
	
	/* Prevent copying into current capture buffer */
	
	return 0;
}

int
Track::invalidate_buffers (bool synchronous)

{
	size_t x;

	if (synchronous) {
		tape.wait_for_butler ();
	} else {
		/* calltree *MUST* have executed wait_for_butler to
		   avoid thread race conditions.
		*/
	}

	for (x = 0; x < nbufs; x++) {
		capture_buffers[x].dirty  = 0;
		capture_buffers[x].needs_flush  = false;
		playback_buffers[x].needs_refill = true;
	}

	playback_buffer = 0;
	playback_samples_left = playback_buffers[playback_buffer].samples;
	playback_ptr = playback_buffers[playback_buffer].buf;

	capture_buffer = 0;
	capture_samples_left = capture_buffers[capture_buffer].samples;
	capture_ptr = capture_buffers[capture_buffer].buf;

	refill_buffer = 0;

	return do_initial_fill ();
}


int
Track::do_initial_fill ()

{
	size_t n;
	
	for (n = 0; n < nbufs; n++) {
		if (do_refill (n)) {
			return -1;
		}
	}

	return 0;
}

int
Track::do_refill (size_t n)

{
	size_t ret;

#ifdef TRACE_TIMING
	unsigned long now, then;

	printf ("%d: filling buffer %u with %u samples from file sample %u\n",
		track_num,
		n,
		playback_buffers[n].samples,
		playback_buffers[n].file_sample);
	
	rdtscl (then);
#endif 	

	playback_buffers[n].file_sample = file_sample;

	if ((ret = sequence_read (edl, playback_buffers[n].buf, 
				  file_sample, playback_buffers[n].samples)) != playback_buffers[n].samples) {

		error << "Track " << track_num
		      << ": cannot read "
		      << playback_buffers[n].samples
		      << " from EDL at byte "
		      << file_sample
		      << " ; return = " 
		      << ret 
		      << " ("
		      << strerror (errno)
		      << ')'
		      << endmsg;
		return -1;
	}

#ifdef TRACE_TIMING	
	rdtscl (now);
	printf ("%d: refill of %u took %.3fmsecs\n", track_num,
		playback_buffers[n].samples,
		(float) (now - then) / 450000.0f);
#endif

	playback_buffers[n].needs_refill = false;
	file_sample += playback_buffers[n].samples;
	return 0;
}	

void
Track::flush_dirty ()

{
	if (capture_buffers[capture_buffer].dirty) {
		capture_buffers[capture_buffer].needs_async_flush = true;
	}
}

int
Track::do_flush (size_t n)

{
	size_t samples_to_write;

	samples_to_write = min (capture_buffers[n].dirty, 
				tape.frames() - (size_t) capture_buffers[n].file_sample);
	
	if (sequence_replace_data (edl, capture_buffers[n].file_sample,
				   capture_buffers[n].buf,
				   samples_to_write)) {

		error << "Track " << track_num
		      << ": cannot write to disk at sample "
		      << capture_buffers[n].file_sample
		      << endmsg;
		return -1;
	}

#ifdef TRACE_TIMING	
	rdtscl (now);
	printf ("%d: flush of %u took %.3fmsecs\n", track_num,
		samples_to_write,
		(float) (now - then) / 450000.0f);
#endif

	capture_buffers[n].needs_flush = false;
	capture_buffers[n].needs_async_flush = false;
	capture_buffers[n].dirty = 0;
	return 0;
}

int
Track::write_behind (bool &flush_required)

{
	int flush_type = 0;
	int flush_buffer;

	flush_required = false;
	
	/* check each buffer to find the first one that needs flushing.
	   start from one behind the current capture buffer, to ensure
	   a reasonable chance at correct ordering.
	*/

	if (capture_buffer == 0) {
		flush_buffer = nbufs - 1;
	} else {
		flush_buffer = capture_buffer - 1;
	}

	for (size_t x = 0; x < nbufs; x++) {
		
		if (capture_buffers[flush_buffer].needs_flush) {
			flush_type = 1;
			break;
		}

		if (capture_buffers[flush_buffer].needs_async_flush) {
			flush_type = 2;
			break;
		}

		flush_buffer++;
		flush_buffer %= nbufs;
	}

	if (flush_type) {
		flush_required = true;

		if (do_flush (flush_buffer)) {
			return -1;
		}

		if (flush_type == 1) { /* whole buffer was flushed */
			
			/* check the next buffer to see if it needs work too */
			
			flush_buffer++;
			flush_buffer %= nbufs;
			
			if (capture_buffers[flush_buffer].needs_flush) {
				return 1;   /* more work needed right away */
			} 
		}
	}

	return 0;
}

int
Track::read_ahead ()

{
	if (playback_buffers[refill_buffer].needs_refill) {
		
		if (do_refill (refill_buffer)) {
			return -1;
		}

		refill_buffer++;
		refill_buffer %= nbufs;
	}

	if (playback_buffers[refill_buffer].needs_refill) {
		return 1;   /* more work needed right away */
	} else {
		return 0;
	}
}

void
Track::set_record_enable (bool yn)

{
	if (yn) {
		atomic_set (&_record_enabled, 1);
		set_monitor_input (true);
	} else {
		atomic_set (&_record_enabled, 0);
		set_monitor_input (false);
	}
}

void
Track::set_monitor_input (bool yn)

{
	if (_input_channel < 0) {
		return;
	}

	if (yn) {
		if (++_monitor_input == 1) {
			mcd->request_monitor_input (_input_channel, true);
			mcd->request_meter_input (_input_channel, true);
		}
	} else if (_monitor_input && --_monitor_input == 0) {
			mcd->request_monitor_input (_input_channel, false);
			mcd->request_meter_input (_input_channel, false);
	}
}

void
Track::reset_monitor_input ()

{
	if (_input_channel < 0) {
		return;
	}

	mcd->request_monitor_input (_input_channel, _monitor_input);
}

void
Track::set_input_channel (int chn)

{
	mcd->adjust_in_channels (_input_channel, chn);
	_input_channel = chn;
}

void
Track::set_output_channel (int chn)

{
	mcd->adjust_out_channels (_output_channel, chn);
	_output_channel = chn;
}

int
Track::wipe ()

{
	piece_t *silence;

	if (silent_buffer == 0) {
		silent_buffer = const_buffer_new (0);
	}

	silence = piece_new (silent_buffer, 0, tape.frames());
	return sequence_replace_piece (edl, 0, silence);
}

Track::Buffer::Buffer ()

{
	needs_refill = false;
	needs_flush = false;
	needs_async_flush = false;
	dirty = 0;
	samples = 0;
	buf = 0;
	file_sample = 0;
}

Track::Buffer::~Buffer ()

{
	if (buf) {
		delete [] buf;
	}
}

void
Track::Buffer::set_size (size_t smples)

{
	if (buf) {
		delete [] buf;
	}

	samples = smples;
	buf = new Sample[samples];
}

void
Track::set_tagname (const string &name)

{
	_tagname = name;
	tag_changed ();

#if 0
	symlink (_path, _tagname + take_number);
#endif
}

int
Track::use_new_take ()

{
	char buf[256];
	size_t take;
	struct stat statbuf;

#define TAKE_LIMIT 10000

	for (take = 1; take <= TAKE_LIMIT; take++) {
		sprintf (buf, "%s/%u.%d",
			 tape.path().c_str(),
			 track_num + 1,
			 take);
		if (stat (buf, &statbuf) < 0) {
			if (errno == ENOENT) {
				break;
			}
		}
	}
	
	if (take > TAKE_LIMIT) {
		error << "Track: more than "
		      << TAKE_LIMIT
		      << " takes exist! Sorry dude, guess again!"
		      << endmsg;
		return -1;
	}

	return use_take (take);
}
