/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: utils.cc,v 1.11 2000/11/04 04:53:33 pbd Exp $
*/

#include <stdio.h> /* for sprintf */
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/asoundlib.h>

#include <pbd/error.h>

#include "utils.h"
#include "riff.h"
#include "seq.h"

using namespace ARDOUR;

void
sample_to_smpte_clock_parts (size_t sample,
			     size_t sample_rate, 
			     float smpte_fps,
			     size_t *hrs_p,
			     size_t *mins_p,
			     size_t *secs_p,
			     size_t *frames_p)

{
	float secs;
	float computed_secs;
	float approx_secs;
	size_t hrs;
	size_t mins;
	
	approx_secs = ((float) sample / sample_rate);
	secs = approx_secs;
	hrs =  (size_t) floor (secs / 3600.0);
	secs -= (hrs * 3600.0);
	mins = (size_t) floor (secs / 60.0);
	secs -= (mins * 60.0);

	computed_secs = (hrs * 3600.0) + (mins * 60.0) + floor (secs);

	*frames_p = (size_t) ((approx_secs - computed_secs) * smpte_fps);
	*secs_p = (size_t) secs;
	*mins_p = mins;
	*hrs_p = hrs;

	return;
}

char *
sample_to_smpte_clock (size_t sample, char *buf, 
		       size_t sample_rate, 
		       float smpte_fps)

{
	float secs;
	float computed_secs;
	float approx_secs;
	size_t hrs;
	size_t mins;
	size_t frames;
	
	approx_secs = ((float) sample / sample_rate);
	secs = approx_secs;
	hrs =  (size_t) floor (secs / 3600.0);
	secs -= (hrs * 3600.0);
	mins = (size_t) floor (secs / 60.0);
	secs -= (mins * 60.0);

	computed_secs = (hrs * 3600.0) + (mins * 60.0) + floor (secs);
	frames = (size_t) ((approx_secs - computed_secs) * smpte_fps);

	sprintf (buf, "%02d:%02d:%02d:%02d",
		 hrs, mins, (size_t) secs, frames);

	return buf;
}


void
elapsed_time_to_str (char *buf, size_t seconds)

{
	size_t days;
	size_t hours;
	size_t minutes;
	size_t s;

	s = seconds;
	days = s / (3600 * 24);
	s -= (days * 3600 * 24);
	hours = s / 3600;
	s -= (hours * 3600);
	minutes = s / 60;
	s -= minutes * 60;
	
	if (days) {
		sprintf (buf, "%u day%s %u hour%s", 
			 days, 
			 days > 1 ? "s" : "",
			 hours,
			 hours > 1 ? "s" : "");
	} else if (hours) {
		sprintf (buf, "%u hour%s %u minute%s", 
			 hours, 
			 hours > 1 ? "s" : "",
			 minutes,
			 minutes > 1 ? "s" : "");
	} else if (minutes) {
		sprintf (buf, "%u minute%s", 
			 minutes,
			 minutes > 1 ? "s" : "");
	} else if (s) {
		sprintf (buf, "%u second%s", 
			 seconds,
			 seconds > 1 ? "s" : "");
	} else {
		sprintf (buf, "no time");
	}
}

char *
format_edl_operation (sequence_op_t *op, char *buf,
		      size_t sample_rate, float smpte_fps,
		      size_t sample_bytes,
		      bool smpte)
		      

{
	gchar *id = NULL;
	buffer_t *buffer;
	BufferType buftype;

	if (smpte) {
		
		/* smpte time units */
		
		char time1[32];
		char time2[32];
		
		switch (op->type) {
		case SequenceInsert:
			sample_to_smpte_clock (op->start/sample_bytes,
					       time1,
					       sample_rate,
					       smpte_fps);
			sample_to_smpte_clock ((op->start+op->cnt)/sample_bytes,
					       time2,
					       sample_rate,
					       smpte_fps);

			buffer = op->piece->buffer;
			buftype = buffer->type (buffer);
			id = buffer->id (buffer);

			switch (buftype){ 
			case FileBuffer:
				sprintf (buf, "insert %s+%u at %s-%s",
					 id, op->piece->start, time1, time2);
				break;
			case ConstBuffer:
				/* special case handling of zero */
				if (strcmp (id, "0") == 0) {
					sprintf (buf, "insert silence at %s-%s",
						 time2, time1);
				} else {
					sprintf (buf, "insert %s at %s-%s",
						 id, time2, time1);
				}
				break;
			}
			break;

		case SequenceReplace:
			sample_to_smpte_clock (op->start/sample_bytes,
					       time1,
					       sample_rate,
					       smpte_fps);
			sample_to_smpte_clock ((op->start + op->cnt)/sample_bytes,
					       time2,
					       sample_rate,
					       smpte_fps);

			buffer = op->piece->buffer;
			buftype = buffer->type (buffer);
			id = buffer->id (buffer);

			switch (buftype) {
			case FileBuffer:
				sprintf (buf, "replace %s-%s with %s+%u",
					 time1, time2, id, op->piece->start);
				break;
			case ConstBuffer:
				/* special case handling of zero */
				if (strcmp (id, "0") == 0) {
					sprintf (buf, "replace %s-%s with silence", time2, time1);
				} else {
					sprintf (buf, "replace %s-%s with %s", time2, time1, id);
				}
				break;
			}
			break;

		case SequenceDelete:
			sample_to_smpte_clock (op->start/sample_bytes,
					       time1,
					       sample_rate,
					       smpte_fps);
			sample_to_smpte_clock ((op->start + op->cnt)/sample_bytes,
					       time2,
					       sample_rate,
					       smpte_fps);
			sprintf (buf, "delete %s-%s", time1, time2);
			break;

		case SequenceSetGain:
			sprintf (buf, "set gain to %.3f", op->gain);
			break;
		}

	} else {

		/* Units of frames */

		buffer = op->piece->buffer;

		switch (op->type) {
		case SequenceInsert:
			sprintf (buf, "insert %s, offset = %u at %u for %u frames",
				 (id = buffer->id (buffer)), 
				 op->piece->start,
				 op->cnt, op->start);
			break;
		case SequenceReplace:
			sprintf (buf, "replace %s offset = %u at %u for %u frames",
				 (id = buffer->id (buffer)),
				 op->piece->start,
				 op->cnt, op->start);
			break;
		case SequenceDelete:
			sprintf (buf, "delete %u frames at %u", op->cnt, op->start);
			break;

		case SequenceSetGain:
			sprintf (buf, "set gain to %.3f", op->gain);
			break;
		}
	}

	if (id) {
		g_free (id);
	}

	return buf;
}


